/**
 * Copyright (c) 2021 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "poker.h"

void pokerInit(poker2_t *poker) {
  poker->playerCount = 0;
  poker->playerDealer = 0;
  poker->playerSmallBlind = 0;
  poker->playerBigBlind = 0;
  pokerResetRound(poker);
}

void pokerResetRound(poker2_t *poker) {
  uint8_t i;
  poker2player_t *player;

  poker->deckSize = cardWriteDeck(poker->deck);
  poker->graveSize = 0;
  poker->communitySize = 0;
  poker->potCount = 0;
  pokerPotAdd(poker);

  for(i = 0; i < poker->playerCount; i++) {
    player = poker->players + i;
    player->cardCount = 0;
    player->currentBet = 0;
    player->state &= ~(
      POKER_PLAYER_STATE_FOLDED |
      POKER_PLAYER_STATE_HAS_BET_THIS_ROUND
    );
  }
}

void pokerNewDealer(poker2_t *poker) {
  uint8_t i, j, k;
  poker2player_t *player;
  bool foundDealer;
  bool foundSmall;

  foundDealer = false;
  foundSmall = false;

  j = poker->playerDealer + 1;
  for(i = 0; i < poker->playerCount; i++) {
    k = (j + i) % poker->playerCount;
    player = poker->players + k;
    if(player->state & POKER_PLAYER_STATE_OUT) continue;
    if(!foundDealer) {
      poker->playerDealer = k;
      foundDealer = true;
    } else if(!foundSmall) {
      poker->playerSmallBlind = k;
      foundSmall = true;
    } else {
      poker->playerBigBlind = k;
      break;
    }
  }
}

void pokerTakeBlinds(poker2_t *poker, int32_t small, int32_t big) {
  pokerPlayerBet(poker, poker->playerSmallBlind, small);
  pokerPlayerBet(poker, poker->playerBigBlind, big);
}

int32_t pokerGetCallValue(poker2_t *poker) {
  uint8_t i;
  int32_t call;
  call = 0;
  for(i = 0; i < poker->playerCount; i++) {
    call = mathMax(call, poker->players[i].currentBet);
  }
  return call;
}

// Pot functions
uint8_t pokerPotAdd(poker2_t *poker) {
  poker2pot_t *pot;
  uint8_t i = poker->potCount++;
  pot = poker->pots + i;
  pot->chips = 0;
  pot->playerCount = 0;
  return i;
}

void pokerPotAddPlayer(poker2pot_t *pot, uint8_t playerIndex) {
  if(arrayContains(
    sizeof(uint8_t), pot->players, pot->playerCount, &playerIndex
  )) return;
  pot->players[pot->playerCount++] = playerIndex;
}

// Dealer Functions
void pokerTurn(poker2_t *poker, uint8_t count) {
  uint8_t i;
  for(i = 0; i < count; i++) {
    cardDeal(
      poker->deck, &poker->deckSize, poker->community, &poker->communitySize
    );
  }
}

void pokerBurn(poker2_t *poker, uint8_t count) {
  uint8_t i;
  for(i = 0; i < count; i++) {
    cardDeal(poker->deck, &poker->deckSize, poker->grave, &poker->graveSize);
  }
}

// Player Functions
uint8_t pokerPlayerAdd(poker2_t *poker) {
  poker2player_t *player;
  uint8_t i = poker->playerCount++;
  
  player = poker->players + i;
  player->cardCount = 0;
  player->chips = 0;
  player->currentBet = 0;
  player->state = POKER_PLAYER_STATE_OUT;
  
  return i;
}

void pokerPlayerDeal(poker2_t *poker, poker2player_t *player, uint8_t count) {
  uint8_t i;
  for(i = 0; i < count; i++) {
    cardDeal(poker->deck, &poker->deckSize, player->cards, &player->cardCount);
  }
}

void pokerPlayerChipsAdd(poker2player_t *player, int32_t chips) {
  player->chips += chips;
  if(player->chips > 0) {
    player->state = flagOff(player->state, POKER_PLAYER_STATE_OUT);
  }
}

void pokerPlayerDealAll(poker2_t *poker, uint8_t count) {
  uint8_t i, j;
  poker2player_t *player;

  for(j = 0; j < count; j++) {
    for(i = 0; i < poker->playerCount; i++) {
      player = poker->players + i;
      
      // Can't deal to a player who is folded or out
      if(player->state & (POKER_PLAYER_STATE_FOLDED|POKER_PLAYER_STATE_OUT)) {
        continue;
      }

      pokerPlayerDeal(poker, player, 1);
    }
  }
}

bool pokerPlayerDoesNeedToBetThisRound(poker2_t *poker, uint8_t playerIndex) {
  poker2player_t *player;
  player = poker->players + playerIndex;
  if(player->state & POKER_PLAYER_STATE_FOLDED) return false;
  if(!(player->state & POKER_PLAYER_STATE_HAS_BET_THIS_ROUND)) return true;
  if(player->currentBet < pokerGetCallValue(poker)) return true;
  return false;
}

uint8_t pokerPlayerGetRemainingBetter(poker2_t *poker) {
  uint8_t i, j;
  poker2player_t *player;

  for(i = 0; i < poker->playerCount; i++) {
    j = (i + poker->playerBigBlind + 1) % poker->playerCount;
    if(pokerPlayerDoesNeedToBetThisRound(poker, j)) return j;
  }
  return 0xFF;
}

uint8_t pokerPlayerGetNextBetter(poker2_t *poker, uint8_t current) {
  uint8_t i, j;
  poker2player_t *player;

  for(i = 0; i < poker->playerCount; i++) {
    j = (i + current + 1) % poker->playerCount;
    if(pokerPlayerDoesNeedToBetThisRound(poker, j)) return j;
  }
  return 0xFF;
}

int32_t pokerPlayerGetCallBet(poker2_t *poker, poker2player_t *player) {
  return pokerGetCallValue(poker) - player->currentBet;
}

// Betting
void pokerPlayerBetPot(
  poker2_t *poker, poker2pot_t *pot, uint8_t playerIndex, int32_t chips
) {
  poker2player_t *player;
  player = poker->players + playerIndex;
  player->chips -= chips;
  player->currentBet += chips;
  pot->chips += chips;
  player->state |= POKER_PLAYER_STATE_HAS_BET_THIS_ROUND;
  pokerPotAddPlayer(pot, playerIndex);
}

void pokerPlayerBet(poker2_t *poker, uint8_t playerIndex, int32_t chips) {
  pokerPlayerBetPot(
    poker, poker->pots + (poker->potCount - 1), playerIndex, chips
  );
}

poker2turn_t pokerTurnFold(poker2_t *poker, uint8_t player) {
  return (poker2turn_t){
    .chips = 0,
    .confidence = 1,
    .type = POKER_TURN_TYPE_FOLD
  };
}

poker2turn_t pokerTurnBet(poker2_t *poker, uint8_t playerIndex, int32_t chips) {
  poker2turn_t turn;
  poker2player_t *player;
  int32_t i;
  player = poker->players + playerIndex;
  turn.confidence = 1;

  if(chips == 0) {
    turn.type = POKER_TURN_TYPE_CHECK;
    turn.chips = 0;
  } else if(player->chips <= chips) {
    turn.chips = player->chips;
    turn.type = POKER_TURN_TYPE_ALL_IN;
  } else {
    turn.chips = chips;
    turn.type = POKER_TURN_TYPE_BET;
    i = pokerGetCallValue(poker);

    if(chips == (i - player->currentBet)) {
      turn.type = POKER_TURN_TYPE_CALL;
    }
  }

  return turn;
}